import os


class Plugin():

    def __init__(self, prefix=''):
        self.name = self.__class__.__name__
        self.miner = None
        self.prefix = prefix

    def before_hook(self, hook_name, payload):
        return True

    def set_miner(self, miner):
        self.miner = miner

    def notify(self, message, _type='info'):
        message = f"[{self.name}] {message}"
        self.miner.notify(message, _type)

    def __getattr__(self, key):
        if self.miner is None or not hasattr(self.miner, key):
            raise AttributeError(key)
        return getattr(self.miner, key)

    def print_txt(self, printable, name):
        with open(self.plugin_file(f'{name}.txt'), 'a') as f:
            print(f'================ Epoch {self.current_epoch} ================\n', file=f)
            print(printable, file=f)
            print("\n\n", file=f)

    @property
    def plugin_dir(self):
        if hasattr(self, '_plugin_dir'):
            return getattr(self, '_plugin_dir')

        plugin_dir = os.path.join(self.code_dir, self.__class__.__name__)
        try:
            os.mkdir(plugin_dir)
        except FileExistsError:
            pass
        self._plugin_dir = plugin_dir
        return self._plugin_dir

    def plugin_file(self, name):
        return os.path.join(self.plugin_dir, name)

    def create_sheet_column(self, key, name):
        self.miner.create_sheet_column(f'{self.prefix}{key}', f'{self.prefix}{name}')

    def update_sheet(self, key, value):
        self.miner.update_sheet(f'{self.prefix}{key}', value)

    def scalars(self, values, graph):
        return self.miner.drawer.scalars(self.miner.current_epoch, values, f'{self.prefix}{graph}')



import logging
import math
import os
import time
from datetime import datetime
from pathlib import Path

import torch
import tqdm
from IPython.core.display import HTML, display

from . import drawers


class Miner(object):
    """The heart of minetorch

    Args:
        alchemistic_directory (string):
            The directory which minetorch will use to store everything in
        model (torch.nn.Module):
            Pytorch model optimizer (torch.optim.Optimizer): Pytorch optimizer
        loss_func (function):
            A special hook function to compute loss, the function receive 2 variable:
            * Miner: the miner object
            * Data: Batch data been yield by the loader
            return value of the hook function should be a float number of the loss
        code (str, optional):
            Defaults to "geass". It's a code name of one
            attempt. Assume one is doing kaggle competition and will try
            different models, parameters, optimizers... To keep results of every
            attempt, one should change the code name before tweaking things.
        train_dataloader (torch.utils.data.DataLoader):
            Pytorch dataloader
        val_dataloader (torch.utils.data.DataLoader, optional):
            Defaults to None, if no validation dataloader is provided, will skip validation
        resume (bool, optional):
            Defaults to True. Resume from last training, could be:
            * True: resume from the very last epochs
            * String: resume from the specified epochs
                          etc. `34`, `68` `best`
        eval_stride (int, optional):
            Defaults to 1. Validate every `eval_stride` epochs
        persist_stride (int, optional):
            Defaults to 1.
            Save model every `persist_stride` epochs
        drawer (minetorch.Drawer or string, optional):
            Defaults to matplotlib.
            If provide, Miner will draw training loss and validation loss
            curves, could be `tensorboard` or self implemented Drawer object
        hooks (dict, optional):
            Defaults to {}. Define hook functions.
        max_epochs ([type], optional):
            Defaults to None. How many epochs to train, None means unlimited.
        logging_format ([type], optional):
            Defaults to None. logging format
        trival ([Boolean], optional):
            Defaults to False. If true, both training and validation
            process will be breaked in 10 iterations
        plugins (list, optional):
            Defaults to []. This is actually a collection of `hooks`, do not set
            hooks and plugins the same time.
        forward (function, optional):
            custom forward function.
        verbose (boolean, optional):
            los loss of every iteration
    """

    def __init__(
            self, alchemistic_directory, model, optimizer, loss_func,
            code="geass", train_dataloader=None, val_dataloader=None,
            resume=True, eval_stride=1, persist_stride=1, gpu=True,
            drawer='matplotlib', hooks={}, max_epochs=None, statable={},
            logging_format=None, trival=False, in_notebook=False, plugins=[],
            logger=None, sheet=None, accumulated_iter=1, ignore_optimizer_resume=False,
            forward=None, verbose=False, amp=False):
        self.alchemistic_directory = alchemistic_directory
        self.code = code
        if trival:
            self.code = f'trival_{code}'
        self.create_dirs()
        self.gpu = gpu
        self.devices = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.logger = logger
        self.code_dir = os.path.join(alchemistic_directory, self.code)
        if self.logger is None:
            self.set_logging_config(alchemistic_directory, self.code, logging_format)
            self.logger = logging
        self.create_drawer(drawer)
        self.models_dir = os.path.join(alchemistic_directory, self.code, 'models')
        self.in_notebook = in_notebook
        self.statable = statable
        self.accumulated_iter = float(accumulated_iter)
        self.ignore_optimizer_resume = ignore_optimizer_resume

        self.model = model
        self.optimizer = optimizer
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader

        self.loss_func = loss_func

        self.resume = resume
        self.eval_stride = eval_stride
        self.persist_stride = persist_stride
        self.lowest_train_loss = float('inf')
        self.lowest_val_loss = float('inf')
        self.current_epoch = 0
        self.current_train_iteration = 0
        self.current_val_iteration = 0
        self.hook_funcs = hooks
        self.max_epochs = max_epochs
        self.trival = trival
        self.forward_fn = forward
        self.verbose = verbose
        self.amp = amp

        self.sheet = sheet
        if self.sheet:
            self._init_sheet()

        self.plugins = plugins
        for plugin in self.plugins:
            plugin.set_miner(self)

        self._set_tqdm()
        self.call_hook_func('before_init')
        self._check_statable()
        self.init_model()
        if self.sheet:
            self.sheet_progress = dict(
                epoch=0,
                train_percentage='0%',
                val_percentage='0%'
            )
            self.last_flushed_at = 0
            self.sheet.onready()
            self.sheet.flush()
        self.status = 'init'
        self.call_hook_func('after_init')

    def _check_statable(self):
        for name, statable in self.statable.items():
            if not (hasattr(statable, 'state_dict') and hasattr(statable, 'load_state_dict')):
                raise Exception(f'The {name} is not a statable object')

    def _set_tqdm(self):
        if self.in_notebook:
            self.tqdm = tqdm.tqdm_notebook
        else:
            self.tqdm = lambda x: x

    def _init_sheet(self):
        self.sheet.set_miner(self)
        self.sheet.reset_index()
        self.sheet.create_column('code', 'Code')
        self.sheet.create_column('progress', 'Progress')
        self.sheet.create_column('loss', 'Loss')
        self.sheet.update('code', self.code)

    def create_sheet_column(self, key, title):
        if self.sheet is None:
            return
        self.sheet.create_column(key, title)

    def update_sheet(self, key, value):
        if self.sheet is None:
            return
        self.sheet.update(key, value)

    def set_logging_config(self, alchemistic_directory, code, logging_format):
        self.log_dir = os.path.join(alchemistic_directory, code)
        log_file = os.path.join(self.log_dir, 'log.txt')
        logging_format = logging_format if logging_format is not None else \
            '%(levelname)s %(asctime)s %(message)s'
        logging.basicConfig(
            filename=log_file,
            format=logging_format,
            datefmt="%m-%d %H:%M:%S",
            level=logging.INFO
        )

    def create_drawer(self, drawer):
        if drawer == 'tensorboard':
            self.drawer = drawers.TensorboardDrawer(self)
        elif drawer == 'matplotlib':
            self.drawer = drawers.MatplotlibDrawer(self)
        else:
            self.drawer = drawer

    def notebook_output(self, message, _type='info'):
        type_config = {
            'info': ['💬', '#6f818a'],
            'success': ['✅', '#7cb305'],
            'error': ['❌', '#cf1322'],
            'warning': ['⚠️', '#d46b08'],
        }[_type]
        if self.in_notebook:
            display(HTML(
                f'<div style="font-size: 12px; color: {type_config[1]}">'
                f'⏰ {time.strftime("%b %d - %H:%M:%S")} >>> '
                f'{type_config[0]} {message}'
                '</div>'
            ))

    def notebook_divide(self, message):
        if self.in_notebook:
            display(HTML(
                '<div style="display: flex; justify-content: center;">'
                f'<h3 style="color: #7cb305; border-bottom: 4px dashed #91d5ff; padding-bottom: 6px;">{message}</h3>'
                '</div>'
            ))

    def init_model(self):
        """resume from some checkpoint
        """
        if isinstance(self.model, torch.nn.DataParallel):
            raise Exception(
                'Don\'t parallel the model yourself, instead, if the '
                '`gpu` option is true(default), Minetorch will do this for you.'
            )

        if self.resume is True:
            # resume from the newest model
            if self.model_file_path('latest') is not None:
                checkpoint_path = self.model_file_path('latest')
            else:
                checkpoint_path = None
                msg = ('Could not find checkpoint to resume, '
                       'train from scratch')
                self.logger.warning(msg)
                self.notebook_output(msg, _type='warning')
        elif isinstance(self.resume, str):
            checkpoint_path = self.model_file_path(self.resume)
        elif isinstance(self.resume, int):
            checkpoint_path = self.model_file_path(str(self.resume))
        else:
            checkpoint_path = None

        if self.resume is not True and self.resume and checkpoint_path is None:
            # user has specified a none existed model, should raise a error
            raise Exception(f"Could not find model {self.resume}")

        if checkpoint_path is not None:
            msg = f"Start to load checkpoint {checkpoint_path}"
            self.logger.info(msg)
            self.notebook_output(msg)
            checkpoint = torch.load(checkpoint_path)
            self.current_epoch = checkpoint.get('epoch', 0)
            self.current_train_iteration = checkpoint.get('train_iteration', 0)
            self.current_val_iteration = checkpoint.get('val_iteration', 0)
            self.lowest_train_loss = checkpoint.get('lowest_train_loss', 9999)
            self.lowest_val_loss = checkpoint.get('lowest_val_loss', 9999)

            try:
                self.model.load_state_dict(checkpoint['state_dict'], strict=True)
            except Exception as e:
                msg = (f'load checkpoint failed with {e}, the state in the '
                       'checkpoint is not matched with the model, '
                       'try to reload checkpoint with unstrict mode')
                self.logger.warning(msg)
                self.notebook_output(msg)
                self.model.load_state_dict(checkpoint['state_dict'], strict=False)

            if 'optimizer' in checkpoint and not self.ignore_optimizer_resume:
                try:
                    self.optimizer.load_state_dict(checkpoint['optimizer'])
                except Exception as e:
                    msg = (f'load optimizer state failed with {e}, will skip this error and continue, '
                           'stop the process if it is not expected')
                    self.logger.warning(msg)
                    self.notebook_output(msg)

            if (self.drawer is not None) and ('drawer_state' in checkpoint):
                self.drawer.set_state(checkpoint['drawer_state'])

            if 'statable' in checkpoint:
                for name, statable in self.statable.items():
                    if name not in checkpoint['statable']:
                        continue
                    statable.load_state_dict(checkpoint['statable'][name])
            msg = 'checkpoint loaded'
            self.notebook_output(msg, _type='success')
        self.model = self.parallel_model(self.model)

    def parallel_model(self, model):
        if self.gpu:
            gpu_count = torch.cuda.device_count()
            if gpu_count == 0:
                self.notify('no GPU detected, will train on CPU.')
            else:
                self.notify(f'found {gpu_count} GPUs, will use all of them to train')
                devices = list(map(lambda x: f'cuda:{x}', range(gpu_count)))
                model.cuda()
                model = torch.nn.DataParallel(model, devices)
        return model

    def notify(self, message, _type='info'):
        getattr(self.logger, _type)(message)
        self.notebook_output(message, _type)

    def call_hook_func(self, name, **payload):
        if name in self.hook_funcs:
            self.hook_funcs[name](miner=self, **payload)

        for plugin in self.plugins:
            if not plugin.before_hook(name, payload):
                continue
            if hasattr(plugin, name):
                getattr(plugin, name)(**payload)

    def train(self):
        """start to train the model
        """
        while True:
            self.current_epoch += 1
            self.call_hook_func('before_epoch_start', epoch=self.current_epoch)
            self.notebook_divide(f'Epoch {self.current_epoch}')
            self.model.train()
            train_iters = len(self.train_dataloader)

            total_train_loss = 0
            percentage = 0
            total = len(self.train_dataloader)
            self.notebook_output(f'start to train epoch {self.current_epoch}')
            self._update_progress(force=True, epoch=self.current_epoch, train_percentage='0%', val_percentage='0%')
            for index, data in enumerate(self.tqdm(self.train_dataloader)):
                if self.trival is True and index == 10:
                    break
                train_loss = self.run_train_iteration(index, data, train_iters)
                self.tqdm.set_postfix({"train loss": train_loss})
                if int((index + 1) % self.accumulated_iter) == 0:
                    self.optimizer.step()
                    self.optimizer.zero_grad()
                total_train_loss += train_loss
                current_percentage = math.ceil(index / total * 100)
                if current_percentage != percentage:
                    self._update_progress(train_percentage=f'{percentage}%')
                    percentage = current_percentage
            self.optimizer.step()
            self.optimizer.zero_grad()
            self._update_progress(force=True, train_percentage=f'{current_percentage}%')

            total_train_loss = total_train_loss / train_iters
            self.notebook_output(f'training of epoch {self.current_epoch} finished, '
                                 f'loss is {total_train_loss}')

            total_val_loss = 0
            percentage = 0
            total = len(self.val_dataloader)
            if self.val_dataloader is not None:
                val_iters = len(self.val_dataloader)
                with torch.set_grad_enabled(False):
                    self.model.eval()
                    self.notebook_output(f'validate epoch {self.current_epoch}')
                    for index, data in enumerate(self.tqdm(self.val_dataloader)):
                        if self.trival is True and index == 10:
                            break
                        val_loss = self.run_val_iteration(index, data, val_iters)
                        self.tqdm.set_postfix({"val loss": val_loss})
                        total_val_loss += val_loss
                        current_percentage = math.ceil(index / total * 100)
                        if current_percentage != percentage:
                            self._update_progress(val_percentage=f'{percentage}%')
                            percentage = current_percentage
                    self._update_progress(force=True, val_percentage=f'{current_percentage}%')

                total_val_loss = total_val_loss / val_iters
                self.notebook_output(f'validation of epoch {self.current_epoch}'
                                     f'finished, loss is {total_val_loss}')
            if self.drawer is not None:
                png_file = self.drawer.scalars(
                    self.current_epoch, {'train': total_train_loss, 'val': total_val_loss}, 'loss'
                )
                if png_file is not None:
                    self.update_sheet('loss', {'raw': png_file, 'processor': 'upload_image'})

            if total_train_loss < self.lowest_train_loss:
                self.lowest_train_loss = total_train_loss

            if total_val_loss < self.lowest_val_loss:
                message = ('current val loss {} is lower than lowest {}, '
                           'persist this model as best one'.format(
                            total_val_loss, self.lowest_val_loss))
                self.notebook_output(f'{message}', _type='success')
                self.logger.info(message)

                self.lowest_val_loss = total_val_loss
                self.persist('best')
            self.persist('latest')

            if not self.current_epoch % self.persist_stride:
                self.persist('epoch_{}'.format(self.current_epoch))

            if self.max_epochs is not None and self.current_epoch >= self.max_epochs:
                self.call_hook_func('before_quit')
                self.logger.info('exceed max epochs, quit!')
                break

            if self.sheet:
                self.sheet.flush()
            self.call_hook_func(
                'after_epoch_end',
                train_loss=total_train_loss,
                val_loss=total_val_loss,
                epoch=self.current_epoch
            )

    def run_train_iteration(self, index, data, train_iters):
        self.status = 'train'
        self.current_train_iteration += 1
        self.call_hook_func(
            'before_train_iteration_start',
            data=data,
            index=index,
            total_iters=train_iters,
            iteration=self.current_train_iteration)
        if self.amp:
            with torch.cuda.amp.autocast():
                _, loss = self._forward(data)
        else:
            _, loss = self._forward(data)
        seperate_loss = loss / self.accumulated_iter
        seperate_loss.backward()
        loss = loss.detach().cpu().item()
        if self.verbose:
            self.logger.info('[train {}/{}/{}] loss {}'.format(
                self.current_epoch, index, train_iters, loss))

        self.call_hook_func(
            'after_train_iteration_end',
            loss=loss,
            data=data,
            index=index,
            total_iters=train_iters,
            iteration=self.current_train_iteration
        )
        return loss

    def _forward(self, data):
        if self.forward_fn:
            return self.forward_fn(self, data)
        else:
            predict = self.model(data[0].to(self.devices))
            loss = self.loss_func(predict, data[1].to(self.devices))
            return predict, loss

    def run_val_iteration(self, index, data, val_iters):
        self.status = 'val'
        self.current_val_iteration += 1
        self.call_hook_func(
            'before_val_iteration_start',
            data=data,
            index=index,
            total_iters=val_iters,
            iteration=self.current_val_iteration
        )
        predict, loss = self._forward(data)
        loss = loss.detach().cpu().item()
        if self.verbose:
            self.logger.info('[val {}/{}/{}] loss {}'.format(
                self.current_epoch, index, val_iters, loss))
        self.call_hook_func(
            'after_val_iteration_ended',
            predicts=predict,
            loss=loss,
            data=data,
            index=index,
            total_iters=val_iters,
            iteration=self.current_val_iteration
        )
        return loss

    def persist(self, name):
        """save the model to disk
        """
        self.call_hook_func('before_checkpoint_persisted')
        if self.drawer is not None:
            drawer_state = self.drawer.get_state()
        else:
            drawer_state = {}

        if isinstance(self.model, torch.nn.DataParallel):
            model_state_dict = self.model.module.state_dict()
        else:
            model_state_dict = self.model.state_dict()

        state = {
            'state_dict': model_state_dict,
            'optimizer': self.optimizer.state_dict(),
            'epoch': self.current_epoch,
            'train_iteration': self.current_train_iteration,
            'val_iteration': self.current_val_iteration,
            'lowest_train_loss': self.lowest_train_loss,
            'lowest_val_loss': self.lowest_val_loss,
            'drawer_state': drawer_state,
            'statable': {}
        }

        for statable_name, statable in self.statable.items():
            state['statable'][statable_name] = statable.state_dict()

        modelpath = self.standard_model_path(name)
        torch.save(state, modelpath)
        message = f'save checkpoint to {self.standard_model_path(name)}'
        self.logger.info(message)
        self.notebook_output(message)
        self.call_hook_func('after_checkpoint_persisted', modelpath=modelpath)

    def standard_model_path(self, model_name):
        return os.path.join(self.models_dir, f'{model_name}.pth.tar')

    def model_file_path(self, model_name):
        model_name_path = Path(str(model_name))
        models_dir_path = Path(self.models_dir)

        search_paths = [
            model_name_path,
            models_dir_path / model_name_path,
            models_dir_path / f'{model_name}.pth.tar',
            models_dir_path / f'epoch_{model_name}.pth.tar',
        ]

        for path in search_paths:
            if path.is_file():
                return path.resolve()

        return None

    # TODO: implement methods below
    def graceful_stop(self):
        """stop train and exist after this epoch
        """
        pass

    def save_and_stop(self):
        """save the model immediately and stop training
        """
        pass

    def create_dirs(self):
        """Create directories
        """
        self.create_dir('')
        self.create_dir(self.code)
        self.create_dir(self.code, 'models')

    def create_dir(self, *args):
        """Create directory
        """
        current_dir = self.alchemistic_directory
        for dir_name in args:
            current_dir = os.path.join(current_dir, dir_name)
            if not os.path.isdir(current_dir):
                os.mkdir(current_dir)

    def periodly_flush(self, force=False):
        if self.sheet is None:
            return
        now = int(datetime.now().timestamp())
        # flush every 10 seconds
        if not force and now - self.last_flushed_at < 10:
            return
        self.sheet.flush()
        self.last_flushed_at = now

    def _update_progress(self, force=False, **kwargs):
        if self.sheet is None:
            return

        self.sheet_progress.update(kwargs)
        progress = f"""
         epoch:  {self.sheet_progress.get('epoch')}
train progress:  {self.sheet_progress.get('train_percentage')}
  val progress:  {self.sheet_progress.get('val_percentage')}
"""
        self.sheet.update('progress', progress)
        self.periodly_flush(force)



import functools
import logging
from concurrent.futures import ThreadPoolExecutor

from google.oauth2 import service_account
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from googleapiclient.http import MediaFileUpload

from .plugin import Plugin

pool = ThreadPoolExecutor(1)
logger = logging.getLogger(__name__)


def _async(fn):

    @functools.wraps(fn)
    def _func(*args, **kwargs):
        def _inner(*args, **kwargs):
            try:
                fn(*args, **kwargs)
            except Exception as e:
                logger.warn(f'error occured while handle task {e}')
                raise e
        return pool.submit(_inner, *args, **kwargs)

    return _func


class ColumnNotExistsError(Exception):
    pass


class MinetorchSpreadsheet(Plugin):

    def __init__(self):
        super().__init__()
        self.columns = []
        self.cached_row_data = {}

    def _create_experiment_row(self):
        """Create a row for this experiment"""
        raise NotImplementedError

    def _create_end_column_divider(self):
        raise NotImplementedError

    @_async
    def update(self, key, value):
        """Update value for a column"""
        if key not in self.columns:
            raise ColumnNotExistsError
        if not isinstance(value, dict):
            value = {'raw': value}
        self.cached_row_data[key] = value

    def flush(self):
        raise NotImplementedError

    def onready(self):
        """Called after all the columns are created"""
        raise NotImplementedError

    def create_column(self, key, title):
        """Create a column on the sheet."""
        self.columns.append(key)

    @property
    def experiment_row_name(self):
        return f'_row_{self.code}'

    @property
    def title_row_name(self):
        return '_row_title'

    @property
    def end_column_name(self):
        return '_column_end_divider'

    @property
    def banner_row_name(self):
        return '_row_banner'


class GoogleSheet(MinetorchSpreadsheet):

    def __init__(self, sheet_id, service_account_file, meta_prefix='', build_kwargs=None):
        super().__init__()

        if build_kwargs is None:
            build_kwargs = {}

        self.sheet_id = sheet_id
        self.meta_prefix = meta_prefix

        credentials = service_account.Credentials.from_service_account_file(service_account_file)
        service = build('sheets', 'v4', credentials=credentials, **build_kwargs)
        self.sheet = service.spreadsheets()
        self.drive = build('drive', 'v3', credentials=credentials)

        self.drive_folder_id = self._prepare_drive_directory()

    def _meta(self, key):
        return f'{self.meta_prefix}{key}'

    def _num_to_letter(self, num):
        num += 1
        letters = ''
        while num:
            mod = (num - 1) % 26
            letters += chr(mod + 65)
            num = (num - 1) // 26
        return ''.join(reversed(letters))

    def _index_of(self, key):
        search = {
            "dataFilters": [{
                "developerMetadataLookup": {
                    "metadataKey": self._meta(key)
                }
            }]
        }
        result = self.sheet.developerMetadata().search(spreadsheetId=self.sheet_id, body=search).execute()
        if len(result.items()) == 0:
            return False
        else:
            return result['matchedDeveloperMetadata'][0]['developerMetadata']['location']['dimensionRange']['startIndex']

    _exists = _index_of

    @_async
    def reset_index(self):
        self.banner_index = self._create_banner_dimension()
        self.title_index = self._create_title_dimension()
        self.endcol_index = self._create_end_column_divider()
        self.experiment_row_index = self._insert_dimension(self.experiment_row_name, self.title_index + 1, 'ROWS')

    @_async
    def prepare(self):
        self.reset_index()
        self.update('code', self.code)

    @property
    def dark_bg(self):
        return {
            "red": 0.10980392156862745,
            "green": 0.5686274509803921,
            "blue": 0.6039215686274509
        }

    @property
    def white(self):
        return {
            "red": 1.0,
            "green": 1.0,
            "blue": 1.0
        }

    @property
    def light_bg(self):
        return {
            "red": 0.9411764705882353,
            "green": 1.0,
            "blue": 0.9882352941176471,
        }

    def _create_banner_dimension(self):
        return self._insert_dimension(self.banner_row_name, 0, 'ROWS')

    def _create_title_dimension(self):
        return self._insert_dimension(self.title_row_name, 1, 'ROWS')

    def _create_end_column_divider(self):
        icol = self._insert_dimension(self.end_column_name, 0, 'COLUMNS')
        requests = [{
            'repeatCell': {
                "range": {
                    "sheetId": 0,
                    "startColumnIndex": icol,
                    "endColumnIndex": icol + 1,
                    "startRowIndex": self.title_index,
                    "endRowIndex": self.title_index + 1
                },
                "cell": {
                    "userEnteredFormat": {
                        "backgroundColor": self.dark_bg,
                        "textFormat": {
                            "foregroundColor": self.white,
                            "bold": False,
                            "fontSize": 12,
                        },
                        "horizontalAlignment": "CENTER",
                        "verticalAlignment": "MIDDLE"
                    }
                },
                "fields": "*",
            }
        }, {
            'updateDimensionProperties': {
                "range": {
                    "sheetId": 0,
                    "dimension": 'COLUMNS',
                    "startIndex": icol,
                    "endIndex": icol + 1,
                },
                "properties": {
                    "hiddenByUser": True,
                },
                "fields": 'hiddenByUser',
            }
        }]

        self.sheet.batchUpdate(spreadsheetId=self.sheet_id, body={"requests": requests}).execute()
        return icol

    def _insert_dimension(self, row_name, index, dim, extra_request=None):
        result = self._exists(row_name)
        if result is not False:
            return result

        if extra_request is None:
            extra_request = []

        create_row_request = {
            "insertDimension": {
                "range": {
                    "sheetId": 0,
                    "dimension": dim,
                    "startIndex": index,
                    "endIndex": index + 1
                },
                "inheritFromBefore": False
            }
        }

        assign_name_request = {
            "createDeveloperMetadata": {
                "developerMetadata": {
                    "metadataKey": self._meta(row_name),
                    "metadataValue": self._meta(row_name),
                    "location": {
                        "dimensionRange": {
                            "sheetId": 0,
                            "dimension": dim,
                            "startIndex": index,
                            "endIndex": index + 1
                        }
                    },
                    "visibility": "DOCUMENT"
                }
            }
        }

        body = {
            "requests": [
                create_row_request,
                assign_name_request,
                *extra_request
            ],
        }

        self.sheet.batchUpdate(spreadsheetId=self.sheet_id, body=body).execute()
        return self._exists(row_name)

    @_async
    def onready(self):
        banner = """😀 Thanks for using Minetorch, if you found it's useful, please considering star the project at https://github.com/minetorch/minetorch.

👇 The cells generated by Minetorch should not be manually modified, but you can still change the size of them.

👉 The other cells are free to edit, Minetorch will never override them.

🧶 Happy tweaking !
"""
        icol_end = self._index_of(self.end_column_name)
        icol_start = self._index_of(self.end_column_name) - len(self.columns)
        merge_cells = {
            "mergeCells": {
                "range": {
                    "sheetId": 0,
                    "startRowIndex": self.banner_index,
                    "endRowIndex": self.banner_index + 1,
                    "startColumnIndex": icol_start,
                    "endColumnIndex": icol_end
                },
                "mergeType": "MERGE_ALL"
            }
        }

        change_cell = {
            "repeatCell": {
                "range": {
                    "sheetId": 0,
                    "startRowIndex": self.banner_index,
                    "endRowIndex": self.banner_index + 1,
                    "startColumnIndex": icol_start,
                    "endColumnIndex": icol_end
                },
                "cell": {
                    "userEnteredFormat": {
                        "wrapStrategy": "WRAP"
                    }
                },
                "fields": "*"
            }
        }

        auto_resize = {
            "autoResizeDimensions": {
                "dimensions": {
                    "sheetId": 0,
                    "dimension": "ROWS",
                    "startIndex": self.banner_index,
                    "endIndex": self.banner_index + 1
                }
            }
        }

        body = {
            "requests": [
                merge_cells,
                change_cell,
                auto_resize
            ]
        }

        self.sheet.batchUpdate(spreadsheetId=self.sheet_id, body=body).execute()
        self._update_cells(f'{self._num_to_letter(icol_start)}{self.banner_index + 1}', [banner])

    @_async
    def create_column(self, key, title, size=None):
        super().create_column(key, title)
        col_index = self._index_of(self.end_column_name)

        change_cell_request = {
            "repeatCell": {
                "range": {
                    "sheetId": 0,
                    "startColumnIndex": col_index,
                    "endColumnIndex": col_index + 1,
                    "startRowIndex": self.experiment_row_index,
                    "endRowIndex": self.experiment_row_index + 1
                },
                "cell": {
                    "userEnteredFormat": {
                        "backgroundColor": self.light_bg,
                        "horizontalAlignment": "CENTER",
                        "verticalAlignment": "MIDDLE"
                    }
                },
                "fields": "*",
            },
        }

        assign_name_request = {
            "createDeveloperMetadata": {
                "developerMetadata": {
                    "metadataKey": self._meta('__minetorch_column__'),
                    "metadataValue": key,
                    "location": {
                        "dimensionRange": {
                            "sheetId": 0,
                            "dimension": 'COLUMNS',
                            "startIndex": col_index,
                            "endIndex": col_index + 1
                        }
                    },
                    "visibility": "DOCUMENT"
                }
            }
        }

        index = self._insert_dimension(key, col_index, 'COLUMNS', [
            change_cell_request,
            assign_name_request
        ])
        self._update_cells(f'{self._num_to_letter(index)}{self.title_index + 1}', [title])

    def _update_cells(self, a1, values):
        value_range = {
            "range": a1,
            "majorDimension": "ROWS",
            "values": [values]
        }
        try:
            self.sheet.values().update(
                spreadsheetId=self.sheet_id,
                range=a1,
                valueInputOption='USER_ENTERED',
                body=value_range).execute()
        except Exception as e:
            self.logger.warn(f'Update sheet failed with {e}')
            return

    @_async
    def flush(self):
        irow = self._index_of(self.experiment_row_name)
        column_indices = self._get_column_indices()
        for key, value in self.cached_row_data.items():
            raw_value = value.get('raw')
            processor = value.get('processor')
            if processor is None:
                value = raw_value
            else:
                value = getattr(self, f'_process_{processor}')(key, raw_value)
            icol = column_indices[key]
            self._update_cells(f'{self._num_to_letter(icol)}{irow + 1}', [value])
        self.cached_row_data = {}

    def _process_upload_image(self, key, value, retry=True):
        try:
            image_id = self._upload_drive_image(key, value)
            return f'=IMAGE("https://drive.google.com/uc?export=view&id={image_id}", 2)'
        except HttpError as e:
            if retry:
                return self._process_upload_image(key, value)
            raise e

    def _process_repr(self, key, value):
        return repr(value)

    def _get_column_indices(self):
        search = {
            "dataFilters": [{
                "developerMetadataLookup": {
                    "metadataKey": self._meta('__minetorch_column__')
                }
            }]
        }
        r = self.sheet.developerMetadata().search(spreadsheetId=self.sheet_id, body=search).execute()
        result = {}
        for item in r['matchedDeveloperMetadata']:
            column_key = item['developerMetadata']['metadataValue']
            index = item['developerMetadata']['location']['dimensionRange']['startIndex']
            result[column_key] = index
        return result

    def _upload_drive_image(self, key, value, retry=True):
        try:
            file_metadata = {'name': key, 'parents': [self.drive_folder_id]}
            media = MediaFileUpload(value, mimetype='image/png')
            file = self.drive.files().create(body=file_metadata, media_body=media, fields='id').execute()
            return file.get('id')
        except HttpError as e:
            if not retry:
                raise e
            self.drive_folder_id = self._prepare_drive_directory()
            self._upload_drive_image(key, value, retry=False)

    def _prepare_drive_directory(self):
        try:
            result = self.drive.files().list(q="name='minetorch_assets'", fields='files(id)').execute()
            dir_id = result['files'][0]['id']
        except IndexError:
            file_metadata = {
                'name': 'minetorch_assets',
                'mimeType': 'application/vnd.google-apps.folder'
            }
            file = self.drive.files().create(body=file_metadata, fields='id').execute()
            dir_id = file.get('id')

        self.drive.permissions().create(fileId=dir_id, body={"role": "writer", "type": "anyone"}).execute()
        return dir_id


"""
if __name__ == '__main__':
    from .miner import Miner
    google_sheet = GoogleSheet('1pxkEJ7h3gH4LNleHSnhHA2jFn-lMbcPpVqgQCh_dF8I', './quickstart.json', 'pre_295')
    google_sheet.code = 'test11'

    google_sheet.reset_index()
    google_sheet.create_column('code', 'Code of the experiment')
    google_sheet.create_column('loss', 'Loss')
    google_sheet.create_column('confusion_matrix', 'Confusion Matrix')
    google_sheet.create_column('some_image', 'Some Image')

    google_sheet.update('code', google_sheet.code)
    google_sheet.flush()

    import numpy as np
    import time

    res = np.random.rand(2, 2)

    google_sheet.update('some_image', {'raw': './examples/alchemistic_directory/geass/graphs/accuracy.png', 'processor': 'upload_image'})
    google_sheet.update('loss', 0.9)
    google_sheet.update('confusion_matrix', repr(res))
    google_sheet.flush()
    google_sheet.onready()
"""



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='Minetorch',
    description='A tools collection for pytorch users',
    version='0.6.8',
    packages=setuptools.find_packages(),
    include_package_data=True,
    url="https://github.com/minetorch/minetorch",
    author='louis',
    author_email='chenglu.she@gmail.com',
    keywords='pytorch minetorch',
    install_requires=[
        'tensorboardX',
        'torch',
        'matplotlib',
        'albumentations',
        'seaborn',
        'google-api-python-client',
        'google-auth-httplib2',
        'google-auth-oauthlib'
    ],
    entry_points={
        'console_scripts': ['minetorch=minetorch.command_line:main'],
    }
)



import torch
import numpy as np


def seed_everything(seed):
    torch.manual_seed(seed)
    # may lead to bad performance
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    np.random.seed(seed)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sn
from sklearn.metrics import (classification_report, cohen_kappa_score,
                             confusion_matrix)

from .plugin import Plugin


class MultiClassesClassificationMetricWithLogic(Plugin):
    """MultiClassesClassificationMetric
    This can be used directly if your loss function is torch.nn.CrossEntropy
    """

    def __init__(
            self, accuracy=True, confusion_matrix=True, kappa_score=True,
            plot_confusion_matrix=True, classification_report=True, sheet_key_prefix=''):

        super().__init__(sheet_key_prefix)
        self.accuracy = accuracy
        self.confusion_matrix = confusion_matrix
        self.kappa_score = kappa_score
        self.plot_confusion_matrix = plot_confusion_matrix
        self.classification_report = classification_report
        self.sheet_key_prefix = sheet_key_prefix

    def before_init(self):
        self.create_sheet_column('latest_confusion_matrix', 'Latest Confusion Matrix')
        self.create_sheet_column('kappa_score', 'Kappa Score')
        self.create_sheet_column('accuracy', 'Accuracy')

    def before_epoch_start(self, epoch):
        self.predicts = np.array([]).astype(np.float)
        self.targets = np.array([]).astype(np.int)

    def after_val_iteration_ended(self, predicts, data, **ignore):
        predicts = np.argmax(predicts.detach().cpu().numpy(), axis=1)

        predicts = predicts.reshape([-1])
        targets = data[1].cpu().numpy().reshape([-1])

        self.predicts = np.concatenate((self.predicts, predicts))
        self.targets = np.concatenate((self.targets, targets))

    def after_epoch_end(self, val_loss, **ignore):
        self.accuracy and self._accuracy()
        self.confusion_matrix and self._confusion_matrix()
        self.kappa_score and self._kappa_score()
        self.classification_report and self._classification_report()
        self.plot_confusion_matrix and self._plot_confusion_matrix(val_loss)

    def _classification_report(self):
        result = classification_report(self.targets, self.predicts)
        self.print_txt(result, 'classification_report')

    def _plot_confusion_matrix(self, val_loss):
        matrix = confusion_matrix(self.targets, self.predicts)
        df_cm = pd.DataFrame(matrix)
        svm = sn.heatmap(df_cm, annot=True, cmap='OrRd', fmt='.3g')
        figure = svm.get_figure()
        if val_loss < self.lowest_val_loss:
            figure.savefig(self.plugin_file('confusion_matrix_epoch_best.png'), facecolor="#F0FFFC")
        figure.savefig(self.plugin_file(f'confusion_matrix_epoch_{self.current_epoch}.png'), facecolor="#F0FFFC")
        figure.savefig(self.plugin_file('confusion_matrix_epoch_latest.png'), facecolor="#F0FFFC")
        plt.clf()

        self.update_sheet('latest_confusion_matrix', {
            'raw': self.plugin_file('confusion_matrix_epoch_latest.png'),
            'processor': 'upload_image'})

    def _accuracy(self):
        png_file = self.scalars(
            {'accuracy': (self.predicts == self.targets).sum() / len(self.predicts)}, 'accuracy'
        )
        if png_file:
            self.update_sheet('accuracy', {'raw': png_file, 'processor': 'upload_image'})

    def _confusion_matrix(self):
        matrix = confusion_matrix(self.targets, self.predicts)
        self.print_txt(matrix, 'confusion_matrix')

    def _kappa_score(self):
        png_file = self.scalars(
            {'kappa_score': cohen_kappa_score(self.targets, self.predicts, weights='quadratic')}, 'kappa_score'
        )

        if png_file:
            self.update_sheet('kappa_score', {'raw': png_file, 'processor': 'upload_image'})
